/*----------------------------------------------------------------------
  File: largeitem.c - largeitem based transaction clustering

  $Log: largeitem.c,v $
  Revision 1.8  2002/02/25 02:22:59  gxd
  Add MaxSessions command line.

  Revision 1.7  2002/01/26 13:24:47  gxd
  add a new parameter: smallweight
  usage:largeitem datafile smallweight outputmode [minsupp[maxcluster[rotatetimes]]]

  Revision 1.6  2002/01/26 07:16:25  gxd
  modify the input parameters
  usage: largeitem datafile output_mode [minsupp[maxcluster[rotate_time]]]

  Revision 1.5  2002/01/16 11:59:58  gxd
  remove the constraint "session length >= 3"

  Revision 1.3  2001/12/02 08:21:52  gxd
  Finished with ComputerDeltaCost. Much faster.

  Revision 1.2  2001/11/30 05:56:01  gxd
  Correct algorithm, but very slow!

  Revision 1.1  2001/11/28 12:25:42  gxd
  Initial revision

----------------------------------------------------------------------*/

#ifndef lint
static char vcid[] = "$Id: largeitem.c,v 1.8 2002/02/25 02:22:59 gxd Exp gxd $";
#endif /* lint */

#include <stdio.h>
#include <stdlib.h>
#include<string.h>
#include <stdarg.h>
#include <assert.h>
#include <time.h>
#include <math.h>

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/

#define BLKSIZE     256         /* block size for enlarging vectors */
#define REMOVED     NULL        /* return value in DecItem() */

typedef struct _node
{
  int iid, support;
  /*  struct _hashelement *hashaddr;*/
  struct _node *prev, *next;
} NODE;

#define MINBULKSIZE           16
#define BULKS                 8192
#define HashEnlargeThreshold  0.2  /* inc bulksize when these percent units allocated */
#define HashShrinkThreshold   0.01  /* dec bulksize when these percent units allocated */
#define NORMALBULKSIZE        16
#define LARGEWEIGHT           10
static int  SMALLWEIGHT = 1;

typedef struct
{
  int bulksize;
  int cnt;
  NODE *bulk[MINBULKSIZE * BULKS];
} HASH;
#define DELETED ((NODE *)1)
#define IsPointer(bulki) ((bulki)>DELETED)

typedef struct {
  HASH *hash;
  NODE *head;
  int  NumSessions;
  int  NumLarge;
  NODE *LargeBegin;
} CF; /* cluster feature */

typedef struct
{
  int size, cnt;
  int ids[1];
} SESSION;

typedef struct {
  int sid;
  int cid;
} SidCid;

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static unsigned long TotalSessions;    /* sesion file format */
static unsigned long TotalItems;       /* first 4 bytes: total sessions */
static struct {                        /* second 4 bytes: total distinct Items */
  unsigned long int sid;               /* then are requests (8 bytes each) */
  unsigned long int iid;               /* strictly ordered by sid */
} request;                             
static char    sessionfilename[256];   /* input sessionfile */
static char    itemfilename[256];      /* the itemfile according to the sessionfile */
static char    clusterfilename[256];   
static char    resultfilename[256];
static char    plotfilename[256];      /* used for gnuplot */
static double  MinSupport=0.6;         /* the value to determine large/small item */
/* restrict the number of clusters, or the clustering speed with be low */
static long    MaxSessions = 1024*1024; /* max sessions to process */
static int     MaxClusters = 50;
static int     RotateTimes = 100;
static int     OutputMode;             /* 0, 1, and 2 */
static CF     *cf = NULL;              /* pointer to array of cfs, dynamic growth */
static int     cfcnt = 0;
static int     cfsize = 0;
static HASH   *LargeHash=NULL, *SmallHash=NULL;
static int    *Result = NULL;          /* pointer to cluster result, dynamic growth */
static int     resultcnt = 0; 
static int     resultsize = 0;
/* control the verbose level of debug information */
#define INFO_INFO  1
#define INFO_WARN  2
#define INFO_INFO2 3
#define INFO_DEBUG 4
#define INFO_DEBUG2 5
static int    VerboseLevel=INFO_INFO;
#define GetMinSupp(N) ((int)((N) * MinSupport + 0.99999))

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

void error(const char *errmsg)  /* -- show error message and exit -- */
{
  fprintf(stderr, "%s\n", errmsg);
  exit(1);
}

void info(int level, char *fmt, ...)
{
  va_list args;
  
  if (level <= VerboseLevel) {
    va_start(args, fmt);
    vfprintf(stdout, fmt, args);
    fflush(stdout);
    va_end(args);
  }
}

/*--------------------------------------------------------------------*/

void DumpOneCluster(int i)
{
  NODE *p;

  assert(i<cfcnt);
  printf("C%d: ", i);
  for (p=cf[i].head; p; p=p->next) {
    if (p!=cf[i].head) printf("->");
    printf("%d:%d", p->iid, p->support);
    if (p == cf[i].LargeBegin) printf("*");
  }
  printf(" (NS: %d NL: %d MinSupp: %d)\n", cf[i].NumSessions, cf[i].NumLarge, 
	 GetMinSupp(cf[i].NumSessions));
}

void DumpCluster()
{
  int i;
  
  for (i=0; i<cfcnt; i++) {
    DumpOneCluster(i);
  }
}

void DumpHash(const char *msg, HASH *h)
{
  int i, n;
  
  printf(msg);
  printf("(");
  for (i=0, n=h->cnt; i<h->bulksize*BULKS; i++) {
    if (IsPointer(h->bulk[i])) {
      if (n!=h->cnt) printf(", ");
      printf("%d:%d(%d)", i, h->bulk[i]->iid, h->bulk[i]->support);
      n--;
    }
  }
  printf("):%d\n", h->cnt);
}

/*--------------------------------------------------------------------*/

int IsInSession(SESSION *s, int iid)
{
  int i;

  assert(s&&s->cnt>=1);
  for (i=0; i<s->cnt; i++) if (s->ids[i]==iid) return 1;
  return 0;
}

void DumpSession(SESSION *s)
{
  int i;

  printf("s->cnt: %d (", s->cnt);
  for (i=0; i<s->cnt;i++) if (i!=0)printf(", %d", s->ids[i]); else printf("%d", s->ids[i]);
  printf(")\n");
  fflush(stdout);
}

/*----------------------------------------------------------------------
  Internal data structure related core functions
----------------------------------------------------------------------*/

HASH *NewHash(int bulksize)
{
  HASH *h;

  assert(bulksize>=1);
  h = (HASH *)calloc(sizeof(HASH)+(bulksize - MINBULKSIZE)*BULKS*sizeof(NODE *), 1);
  if (!h) error("Not enough memory.");
  assert(!h->bulk[bulksize*BULKS-1]);
  h->bulksize=bulksize;
  return h;
}

void AddNewCluster()
{
  void *temp;
  int newsize;

  if (!cf) {
    cf = (CF *)malloc(sizeof(CF)*BLKSIZE);
    if (!cf) error("Not enough memory.");
    cfcnt = 0;
    cfsize = BLKSIZE;
  }
  if (cfcnt == cfsize) {
    newsize = cfsize + ((cfsize < BLKSIZE) ? BLKSIZE : (cfsize >> 1));
    assert(newsize > cfsize);
    temp = (CF *)realloc(cf, sizeof(CF)*newsize);
    if (!temp) error("Not enough memory.");
    cf = temp;
    cfsize = newsize;
  }
  assert(cfcnt < cfsize);
  cf[cfcnt].hash = NewHash(MINBULKSIZE);
  cf[cfcnt].head = NULL;
  cf[cfcnt].LargeBegin = NULL;
  cf[cfcnt].NumLarge = 0;
  cf[cfcnt].NumSessions = 0;
  cfcnt++;
}

void InsertNode(NODE **head, NODE *n, NODE *potentialprev)
{
  NODE *p;
  assert(head && n);
  p = *head;
  if (!p) {
    *head = n;
    n->prev = n->next = NULL;
  } else {
    if ((p->support>n->support) 
	|| ( p->support==n->support && p->iid>n->iid)) {
      *head = p->prev = n;
      n->next = p;
      n->prev = potentialprev;
    } else {
      while (p->next && (p->next->support < n->support 
			 || (p->next->support == n->support
			     && p->next->iid < n->iid ))) p = p->next;
      if (p->next) { 
	p->next->prev = n;
      }
      n->prev = p;
      n->next = p->next;
      p->next = n;
    }
  }
}

#define computehash(iid, n) ((iid) % (n))

void ResizeHash(HASH **hashp, int newbulksize)
{
  HASH *new;
  HASH *old;
  int i, j, start;

  assert(hashp && *hashp);
  old = *hashp;
  info(INFO_DEBUG2, "Rearrange! %d / %d => %d\n", 
	 old->cnt, old->bulksize*BULKS, newbulksize*BULKS);

  assert(newbulksize*BULKS > old->cnt);
  new = NewHash(newbulksize);

  for (i=0;i<old->bulksize*BULKS;i++) {
    if (IsPointer(old->bulk[i])) {
      start=computehash(old->bulk[i]->iid, BULKS) * new->bulksize;
      for (j=start; j<new->bulksize * BULKS && IsPointer(new->bulk[j]); j++);
      if (j==new->bulksize*BULKS) {
	for (j=0; j<start && IsPointer(new->bulk[j]); j++);
      }
      assert(!new->bulk[j]);
      new->bulk[j]=old->bulk[i];
      new->cnt++;
    }
  }
  assert(new->cnt == old->cnt);
  free(old);
  *hashp=new;
}

void EnlargeHash(HASH **hashp)
{
  int newbulksize;

  assert(hashp && *hashp);
  newbulksize = ((*hashp)->bulksize < NORMALBULKSIZE) ? 
    2 * (*hashp)->bulksize : 1.5 * (*hashp)->bulksize;
  ResizeHash(hashp, newbulksize);
}

void ShrinkHash(HASH **hashp)
{
  int newbulksize;

  assert(hashp && *hashp);
  if ((*hashp)->bulksize <= NORMALBULKSIZE) return;

  newbulksize = (*hashp)->bulksize >> 1;
  ResizeHash(hashp, newbulksize);
}

NODE *AddToHash(HASH **hashp, int iid)
{
  HASH *h;
  int i, start, found;
  int FirstDeleted = -1;

  assert(hashp && *hashp);
  if ((*hashp)->cnt / (double)((*hashp)->bulksize*BULKS) 
      > HashEnlargeThreshold) {
    EnlargeHash(hashp); 
  }
  h=*hashp;
  assert(h->cnt < h->bulksize*BULKS);  /* ensure there are empty cells */

  found = 0;
  start = computehash(iid, BULKS) * h->bulksize;
  for (i=start; i<start+h->bulksize && h->bulk[i]; i++) { 
    /* search in the belonged bulk */
    if (IsPointer(h->bulk[i]) && h->bulk[i]->iid == iid) { found = 1; break; }
    else if (h->bulk[i]==DELETED && FirstDeleted == -1) { FirstDeleted = i; }
  }
  if (i>=start+h->bulksize && !found) {
    for (; i<BULKS * h->bulksize && h->bulk[i]; i++) {
      /* search in the remaining bulk */
      if (IsPointer(h->bulk[i]) && h->bulk[i]->iid == iid) { 
	found = 1; break; 
      }
      else if (h->bulk[i]==DELETED && FirstDeleted == -1) { FirstDeleted = i; }
    }
    if (i>=BULKS*h->bulksize && !found) {
      for (i=0; i<start && h->bulk[i]; i++) {
	/* search in the starting bulk */
	if (IsPointer(h->bulk[i]) && h->bulk[i]->iid == iid) {
	  found = 1; break; 
	}
	else if (h->bulk[i]==DELETED && FirstDeleted == -1) { 
	  FirstDeleted = i; 
	}
      }
      if (i>=start) {
	assert(FirstDeleted != -1); /* Otherwise, no room in hash */
	i = FirstDeleted;
      }
    }
  }

  assert(!IsPointer(h->bulk[i]) || h->bulk[i]->iid == iid);

  if (IsPointer(h->bulk[i]) && h->bulk[i]->iid == iid) { /* found */
    h->bulk[i]->support++;
  } else {
    assert(!IsPointer(h->bulk[i]));
    h->bulk[i]=malloc(sizeof(NODE)); 
    if (!h->bulk[i]) error("Not enough memory.");
    h->bulk[i]->iid=iid;
    h->bulk[i]->support=1;
    h->cnt ++;
  }
  return h->bulk[i];
}

int LocateInHash(HASH *h, int iid)
{
  int i, start, found;

  assert(h);
  if (h->cnt==0) return -1; /* can't find anything in an empty hash */
  found = 0;
  start = computehash(iid, BULKS) * h->bulksize;
  for (i=start; i<start+h->bulksize && h->bulk[i]; i++) { 
    /* search in the belonged bulk */
    if (IsPointer(h->bulk[i]) && h->bulk[i]->iid==iid) { found = 1; break; }
  }
  if (i>=start+h->bulksize && !found) {
    for (; i<BULKS * h->bulksize && h->bulk[i]; i++) {
      /* search in the remaining bulk */
      if (IsPointer(h->bulk[i]) && h->bulk[i]->iid==iid) { found = 1; break; }
    }
    if (i>=BULKS*h->bulksize && !found) {
      for (i=0; i<start && h->bulk[i]; i++) {
	/* search in the starting bulk */
	if (IsPointer(h->bulk[i]) && h->bulk[i]->iid==iid) {
	  found = 1; break; 
	}
      }
    }
  }
  if (!found) {
    return (-1);
  }
  assert(found);
  assert(IsPointer(h->bulk[i]));
  return(i);
}

void RemoveFromHash(HASH **hashp, int iid)
{
  HASH *h;
  int i;

  assert(hashp && *hashp);
  h=*hashp;
  assert(h->cnt < h->bulksize*BULKS);  /* ensure there are empty cells */

  i=LocateInHash(h, iid);
  if (i==-1) {
    printf("Error: %d not find in hash.\n", iid);
    DumpHash("Debug dump: ", h);
    DumpHash("Small: ", SmallHash);
    DumpHash("Large: ", LargeHash);
    DumpCluster();
    fflush(stdout);
  }
  assert(i!=-1);
  assert(h->bulk[i]);

  if (--h->bulk[i]->support == 0) {
    free(h->bulk[i]);
    h->bulk[i]=DELETED;
    --h->cnt;
    if ( h->cnt / (double)(h->bulksize*BULKS) < HashShrinkThreshold ) {
      ShrinkHash(hashp); 
    }
  }
}

void AdjustLargeBegin(int clusterid) /* called after each reorder of the dual list */
{
  assert(clusterid<cfcnt && cf[clusterid].LargeBegin);

  if (cf[clusterid].LargeBegin->support>=GetMinSupp(cf[clusterid].NumSessions)) {
    while (cf[clusterid].LargeBegin && cf[clusterid].LargeBegin->prev 
	   && cf[clusterid].LargeBegin->prev->support
	   >=GetMinSupp(cf[clusterid].NumSessions))
      cf[clusterid].LargeBegin = cf[clusterid].LargeBegin->prev;
  } else {
    while (cf[clusterid].LargeBegin && cf[clusterid].LargeBegin->next 
	   && cf[clusterid].LargeBegin->support
	   <GetMinSupp(cf[clusterid].NumSessions))
      cf[clusterid].LargeBegin = cf[clusterid].LargeBegin->next;
  }
}

NODE *IncItem(int clusterid, int iid)
{
  NODE *p, **head;

  info(INFO_DEBUG, "Adding %d to cluster %d ...\n", iid, clusterid);
  assert(clusterid<cfcnt);

  p = AddToHash(&(cf[clusterid].hash), iid);

  if (p->support!=1) {
    if (p->next) {
      if (cf[clusterid].LargeBegin==p) {
	cf[clusterid].LargeBegin = p->next;
      }
      /* detach p from the dual list */
      p->next->prev=p->prev;
      if (!p->prev) {
	cf[clusterid].head=p->next;
	/* insert back */
	InsertNode(&(cf[clusterid].head), p, NULL); 
      } else {
	p->prev->next = p->next;
	/* insert back */
	InsertNode(&(p->prev->next), p, p->prev);
      }
      AdjustLargeBegin(clusterid);
    }
  } else { /* p->support==1, means newly added */
    if (!cf[clusterid].head) {
      cf[clusterid].head=p;
      p->prev = p->next = NULL;
      assert(!cf[clusterid].LargeBegin);
      cf[clusterid].LargeBegin = p;
    } else {
      head = &(cf[clusterid].head);
      InsertNode(head, p, NULL);
      AdjustLargeBegin(clusterid);
    }
  }
  return p;
}

NODE *DecItem(int clusterid, int iid)
{
  NODE *p;
  int i;

  info(INFO_DEBUG, "Removing %d from cluster %d ...\n", iid, clusterid);
  assert(clusterid<cfcnt);

  i = LocateInHash(cf[clusterid].hash, iid);
  assert(i!=-1);

  p = cf[clusterid].hash->bulk[i];

  assert(p&&p->iid==iid);

  /* detach p from the dual list */
  if (p->prev) {
    p->prev->next = p->next;
  } else cf[clusterid].head=p->next;
  if (p->next) {
    p->next->prev = p->prev;
  }
  /* adjust LargeBegin */
  if (cf[clusterid].LargeBegin==p) {
    if (p->next) cf[clusterid].LargeBegin=p->next;
    else if (p->prev) cf[clusterid].LargeBegin=p->prev;
    else cf[clusterid].LargeBegin=NULL;
  }
  if (--p->support>0) {
    /* insert back, maybe modify InsertNode to search by prev */
    InsertNode(&(cf[clusterid].head), p, NULL);
    /* adjust LargeBegin */
    if (!cf[clusterid].LargeBegin) { /* only p in the list */
      cf[clusterid].LargeBegin=p;
    } else {
      AdjustLargeBegin(clusterid);
    }
    return p;
  } else {
    /* free p */
    free(p);
    cf[clusterid].hash->bulk[i]=DELETED;
    cf[clusterid].hash->cnt--;
    return REMOVED;
  }
}

/*--------------------------------------------------------------------*/

void DebugCheckCluster(int clusterid) /* check the cluster data */
{
  NODE *p;
  int cursupport, curiid;
  int MinSupp = GetMinSupp(cf[clusterid].NumSessions);
  
  if (cf[clusterid].head) {
    assert(cf[clusterid].LargeBegin);
    cursupport=cf[clusterid].head->support;
    curiid = cf[clusterid].head->iid;
    for (p=cf[clusterid].head->next; p; p=p->next) {
      assert(p->support>cursupport || (p->support==cursupport && p->iid>curiid));
      cursupport=p->support;
      curiid=p->iid;
    }
    p=cf[clusterid].LargeBegin;
    if (p->support==MinSupp) { /* make sure it points to the first MinSupp node */
      if (p->prev) assert(p->prev->support<MinSupp);
      else assert(cf[clusterid].head == p);
    } else if (p->support<MinSupp) { /* make sure it points to the last elements */
      if (p->next) {DumpCluster(); printf("Cluster %d error!\n", clusterid);}
      assert(!p->next);
    } else { /* > MinSupp, make sure there is no MinSupp node ahead */
      if (p->prev && p->prev->support>=MinSupp) DumpOneCluster(clusterid);
      assert(!p->prev || p->prev->support<MinSupp);
    }
  }
}

void DebugCheckLargeSmall()
{
  /* for all cluster->head: remove small, remove large, add small, add large */
  NODE *p;
  int i;

  for (i=0; i<cfcnt; i++) {
    for (p = cf[i].head; p; p=p->next) {
      if (p->support<GetMinSupp(cf[i].NumSessions)) 
	RemoveFromHash(&SmallHash, p->iid);
      else RemoveFromHash(&LargeHash, p->iid);
    }
  }

  for (i=0; i<cfcnt; i++) {
    for (p = cf[i].head; p; p=p->next) {
      if (p->support<GetMinSupp(cf[i].NumSessions)) 
	AddToHash(&SmallHash, p->iid);
      else AddToHash(&LargeHash, p->iid);
    }
  }
}

/*--------------------------------------------------------------------*/

void AddSession(int clusterid, SESSION *s)
{
  int i;
  int MinSupp, OldMinSupp;
  NODE *p;

  assert(s && s->cnt>0 && cfcnt>clusterid);

  for (i=0;i<s->cnt;i++) {
    p=IncItem(clusterid, s->ids[i]);
    if (VerboseLevel >=INFO_DEBUG) {
      DumpOneCluster(clusterid);
      if (VerboseLevel >=INFO_DEBUG2) {
	DumpHash("Small: ", SmallHash);
	DumpHash("Large: ", LargeHash);
      }
    }
    if (p->support == 1) { /* newly added item, put to SmallHash/LargeHash */
      if (GetMinSupp(cf[clusterid].NumSessions+1)>1) {
	AddToHash(&SmallHash, s->ids[i]); 
	info(INFO_DEBUG2, "S+: %d\n", s->ids[i]);
      } else {
	AddToHash(&LargeHash, s->ids[i]); 
	info(INFO_DEBUG2, "L+: %d\n", s->ids[i]);
	cf[clusterid].NumLarge++;
      }
    }
  } /* end of for */

  if (++cf[clusterid].NumSessions>1) {
    OldMinSupp = GetMinSupp(cf[clusterid].NumSessions-1);
    MinSupp = GetMinSupp(cf[clusterid].NumSessions);
    if (MinSupp == OldMinSupp) {
      for (p = cf[clusterid].LargeBegin; p && p->support == MinSupp; p=p->next) {
	if (IsInSession(s, p->iid)&&p->support>1) {
	  info(INFO_DEBUG2, "S->L: %d\n", p->iid);
	  cf[clusterid].NumLarge++;
	  AddToHash(&LargeHash, p->iid);
	  RemoveFromHash(&SmallHash, p->iid); 
	}
      }
    }
    if (MinSupp == OldMinSupp + 1) {
      for (p = cf[clusterid].LargeBegin; p && p->support == OldMinSupp; p=p->next) {
	if (!IsInSession(s, p->iid)) {
	  info(INFO_DEBUG2, "L->S: %d\n", p->iid);
	  cf[clusterid].NumLarge--;
	  RemoveFromHash(&LargeHash, p->iid);  
	  AddToHash(&SmallHash, p->iid); 
	}
      }
      if (!p) {
	p=cf[clusterid].LargeBegin;
	while (p && p->next && p->next->support < MinSupp) p = p->next;
      }
      cf[clusterid].LargeBegin = p;
    }
  }
  if (VerboseLevel >= INFO_DEBUG2) DebugCheckLargeSmall();
  if (VerboseLevel >= INFO_INFO2) DebugCheckCluster(clusterid);
}

void RemoveSession(int clusterid, SESSION *s)
{
  int i;
  int MinSupp, OldMinSupp;
  NODE *p;

  assert(s && s->cnt>0 && cfcnt>clusterid);

  for (i=0;i<s->cnt;i++) {
    p = DecItem(clusterid, s->ids[i]);
    if (VerboseLevel >=INFO_DEBUG) {
      DumpOneCluster(clusterid);
      if (VerboseLevel >=INFO_DEBUG2) {
	DumpHash("Small: ", SmallHash);
	DumpHash("Large: ", LargeHash);
      }
    }
    if (REMOVED == p) {
      if (GetMinSupp(cf[clusterid].NumSessions)>1) {
	info(INFO_DEBUG2, "S-: %d\n", s->ids[i]);
	RemoveFromHash(&SmallHash, s->ids[i]); 
      } else {
	info(INFO_DEBUG2, "L-: %d\n", s->ids[i]);
	RemoveFromHash(&LargeHash, s->ids[i]); 
	cf[clusterid].NumLarge--;
      }
    }
  }

  if (--cf[clusterid].NumSessions>0) {
    OldMinSupp = GetMinSupp(cf[clusterid].NumSessions+1);
    MinSupp = GetMinSupp(cf[clusterid].NumSessions);
    if (MinSupp == OldMinSupp - 1) { /* Small become large */
      p=cf[clusterid].LargeBegin;
      while (p && p->support >= OldMinSupp) p=p->prev;
      for (; p && p->support == MinSupp; p=p->prev) {
	if (!IsInSession(s, p->iid)) {
	  info(INFO_DEBUG2, "S->L: %d\n", p->iid);
	  cf[clusterid].NumLarge++;
	  AddToHash(&LargeHash, p->iid);
	  RemoveFromHash(&SmallHash, p->iid); 
	}
      }
      if (!p) p = cf[clusterid].head;
      while (p&&p->next&&p->support<MinSupp) p=p->next;
      cf[clusterid].LargeBegin = p;
    }
    if (MinSupp == OldMinSupp) { /* Large become small */
      p = cf[clusterid].LargeBegin;
      while (p && p->support>=MinSupp) p=p->prev;
      for (; p && p->support == MinSupp-1; p=p->prev) {
	if (IsInSession(s, p->iid)) {
	  info(INFO_DEBUG2, "L->S: %d\n", p->iid);
	  cf[clusterid].NumLarge--;
	  RemoveFromHash(&LargeHash, p->iid);	
	  AddToHash(&SmallHash, p->iid); 
	}
      }
    }
  }
  if (VerboseLevel >= INFO_DEBUG2) DebugCheckLargeSmall();
  if (VerboseLevel >= INFO_INFO2) DebugCheckCluster(clusterid);
}

int InSmall(int iid)
{
  if (LocateInHash(SmallHash, iid)!=-1) return 1; else return 0;
}

int InLarge(int iid)
{
  if (LocateInHash(LargeHash, iid)!=-1) return 1; else return 0;
}

int SmallBecomeLarge(int iid)
{
  int j;
  int Cost=0;
  
  j = LocateInHash(SmallHash, iid);
  assert(j!=-1 && SmallHash->bulk[j]->iid == iid);
  if (SmallHash->bulk[j]->support==1) { /* will be removed */
    Cost -= SMALLWEIGHT;
  }
  j = LocateInHash(LargeHash, iid);
  if (j!=-1) Cost += SMALLWEIGHT;
  return Cost;
}

int Cost()
{
  int i, c;

  for (i=0, c=0; i<cfcnt; i++) {
    c += cf[i].NumLarge;
  }
  c = SMALLWEIGHT * SmallHash->cnt + LARGEWEIGHT * (c - LargeHash->cnt);

  return c;
}

/* DeltaMode */
#define DELTA_ADD 0
#define DELTA_REMOVE 1
/*
 * Let C ::= the set of ids whose support = MinSupp
 * Let B ::= the set of ids whose support = MinSupp-1
 * Let S ::= the set of ids in current session
 * The logic:
 * 1) ADD: 
 *  a) if MinSupp will +1, those in C but not in S will become SMALL
 *  b) if MinSupp remains, those in B (if MinSupp==1, in B has no meaning, 
 *     adopt 1.c) and also in S will become LARGE
 *  c) anyway, for those in S that are brand new in the cluster, add them 
       either to the large or to the small. 
 * 2) REMOVE:
 *  a) if MinSupp will -1, those in B but not is S will become LARGE
 *  b) if MinSupp remains and >1, those in C and also in S will 
 *     become SMALL (if MinSupp == 1, adopt 2.c)
 *  c) anyway, for those in S that are removed total from cluster, remove 
 *     them either from small or from large
 */

int ComputeDeltaCost(int DeltaMode, int clusterid, SESSION *s)
{
  NODE *p;
  int i, j, OldMinSupp;
  int DeltaCost=0;
  
  if (!cf[clusterid].head) { /* add to an empty cluster */
    assert(DeltaMode==DELTA_ADD);
    for (i=0;i<s->cnt;i++) DeltaCost+=InLarge(s->ids[i]);
    return DeltaCost * LARGEWEIGHT;
  }
  OldMinSupp = GetMinSupp(cf[clusterid].NumSessions);
  if (DeltaMode==DELTA_ADD) {
    if (GetMinSupp(cf[clusterid].NumSessions+1)>OldMinSupp) { /* (1.a) +1 */
      info(INFO_DEBUG2, "(1.a)\n");
      for(i=0, p=cf[clusterid].LargeBegin; p&&p->support==OldMinSupp; p=p->next) {
	while (i<s->cnt && s->ids[i]<p->iid) i++;
	if (i>=s->cnt || s->ids[i]!=p->iid) { /* in C but not in S, become SMALL */
	  j = LocateInHash(LargeHash, p->iid);
	  assert(j!=-1 && LargeHash->bulk[j]->iid == p->iid);
	  if (LargeHash->bulk[j]->support>1) { /* will not be removed */
	    info(INFO_DEBUG2, "L-(%d), Delta -1\n", p->iid);
	    /* Sum(|Large|) will decrease, but LargeHash will not decrease */
	    DeltaCost -= LARGEWEIGHT;  
	  } else info(INFO_DEBUG2, "L-(%d), Delta 0\n", p->iid);
	  info(INFO_DEBUG2, "S+(%d), Delta %d\n", p->iid, 1-InSmall(p->iid));
	  DeltaCost += (1-InSmall(p->iid)) * SMALLWEIGHT;	  
	}
      }
    } else { /* (1.b) remain */
      info(INFO_DEBUG2, "(1.b)\n");
      if (OldMinSupp!=1) { /* for all those in S and in B */
	for (p=cf[clusterid].LargeBegin; p&&p->support>=OldMinSupp; p=p->prev);
	for (i=s->cnt-1; p&&p->support==OldMinSupp-1; p=p->prev) {
	  while (i>=0&&s->ids[i]>p->iid) i--;
	  if (i<0) break;
	  if (s->ids[i]==p->iid) { /* in S and also in B, become LARGE */
	    info(INFO_DEBUG2, "S->L(%d), Delta+=%d\n", p->iid, SmallBecomeLarge(p->iid));
	    DeltaCost += SmallBecomeLarge(p->iid);
	  }
	}
      }
    }
    /* (1.c) */
    for (i=0; i<s->cnt; i++) {
      if (-1 == LocateInHash(cf[clusterid].hash, s->ids[i])) { /* brand new iid */
	if (GetMinSupp(cf[clusterid].NumSessions+1)>1) { /* add to small */
	  info(INFO_DEBUG2, "->S(%d), Delta+=%d\n", s->ids[i], 1-InSmall(s->ids[i]));
	  DeltaCost += (1-InSmall(s->ids[i])) * SMALLWEIGHT;
	} else { /* add to large */
	  info(INFO_DEBUG2, "->L(%d), Delta+=%d\n", s->ids[i], InLarge(s->ids[i]));
	  DeltaCost += InLarge(s->ids[i]) * LARGEWEIGHT;
	}
      }
    }
  } else {
    assert(DeltaMode==DELTA_REMOVE);
    assert(cf[clusterid].NumSessions>=1);
    if (GetMinSupp(cf[clusterid].NumSessions-1)<OldMinSupp) { /* (2.a) -1 */
      info(INFO_DEBUG2, "(2.a)\n");
      for (p=cf[clusterid].LargeBegin; p&&p->support>=OldMinSupp; p=p->prev);
      for (i=s->cnt-1; p && p->support==OldMinSupp-1; p=p->prev) { 
	/* in B but not in S will become LARGE */
	while (i>=0&&s->ids[i]>p->iid) i--;
	if (i<0 || s->ids[i]!=p->iid) { 
	  info(INFO_DEBUG2, "S->L(%d), Delta+=%d\n", p->iid, SmallBecomeLarge(p->iid));
	  DeltaCost += SmallBecomeLarge(p->iid);
	}
      }
    } else if (OldMinSupp>1) { /* (2.b) same */
      info(INFO_DEBUG2, "(2.b)\n");
      for(i=0, p=cf[clusterid].LargeBegin; p&&p->support==OldMinSupp; p=p->next) {
	while (i<s->cnt && s->ids[i]<p->iid) i++;
	if (i>=s->cnt) break;
	if (s->ids[i]==p->iid) { 
	  /* in C and also in S, become SMALL */
	  j = LocateInHash(LargeHash, p->iid);
	  assert(j!=-1 && LargeHash->bulk[j]->iid == p->iid);
	  if (LargeHash->bulk[j]->support>1) { /* will not be removed */
	    info(INFO_DEBUG2, "L-(%d), Delta -1\n", p->iid);
	    DeltaCost-=LARGEWEIGHT;
	  } else info(INFO_DEBUG2, "L-(%d), Delta 0\n", p->iid);
	  info(INFO_DEBUG2, "S+(%d), Delta %d\n", p->iid, 1-InSmall(p->iid));
	  DeltaCost += (1-InSmall(p->iid)) * SMALLWEIGHT;
	}
      }
    }
    /* (2.c) */
    for (i=0; i<s->cnt; i++) {
      j = LocateInHash(cf[clusterid].hash, s->ids[i]);
      if (j!=-1 && cf[clusterid].hash->bulk[j]->support==1) { /* single iid */
	if (OldMinSupp>1) { /* remove from small */
	  j = LocateInHash(SmallHash, s->ids[i]);
	  assert(j!=-1 && SmallHash->bulk[j]->iid == s->ids[i]);
	  if (SmallHash->bulk[j]->support==1) { /* will not be removed */
	    info(INFO_DEBUG2, "S-(%d), Delta -1\n", s->ids[i]);
	    DeltaCost-=SMALLWEIGHT;
	  } else info(INFO_DEBUG2, "S-(%d), Delta 0\n", s->ids[i]);
	} else { /* remove from large */
	  j = LocateInHash(LargeHash, s->ids[i]);
	  assert(j!=-1 && LargeHash->bulk[j]->iid == s->ids[i]);
	  if (LargeHash->bulk[j]->support>1) { /* will not be removed */
	    info(INFO_DEBUG2, "L-(%d), Delta -1\n", s->ids[i]);
	    DeltaCost-=LARGEWEIGHT;
	  } else info(INFO_DEBUG2, "L-(%d), Delta 0\n", s->ids[i]);
	}
      }
    }
  }

  if (VerboseLevel>=INFO_INFO2) {
    int OldCost;

    OldCost=Cost();
    if (DeltaMode == DELTA_ADD) {
      AddSession(clusterid, s);
      printf("+Cluster %d, DeltaCost: %d, Cost()-OldCost: %d\n", 
	     clusterid, DeltaCost, Cost()-OldCost);
      assert(DeltaCost == Cost()-OldCost);
      RemoveSession(clusterid, s);
    } else {
      RemoveSession(clusterid, s);
      printf("-Cluster %d, DeltaCost: %d, Cost()-OldCost: %d\n", 
	     clusterid, -DeltaCost, -Cost()+OldCost);
      assert(DeltaCost == Cost()-OldCost);
      AddSession(clusterid, s);
    }
  }
  return DeltaCost;
}

/* Result[sid]=clusterid */
/* return 1 if different or new, return 0 if same */
int SetResult(int sid, int clusterid)
{
  int newsize;
  void *temp;

  if (Result && sid < resultcnt) {
    if (Result[sid] != clusterid) {
      Result[sid] = clusterid; 
      return 1;
    }
    return 0;
  } else {
    assert(sid == resultcnt);
    if (resultcnt == resultsize) {
      newsize = resultsize + ((resultsize < BLKSIZE) ? BLKSIZE : (resultsize >> 1));
      assert(newsize > resultsize);
      temp = (int *)realloc(Result, sizeof(int)*newsize);
      if (!temp) error("Not enough memory.");
      Result=temp;
      resultsize=newsize;
    }
    assert(resultcnt < resultsize);
    Result[resultcnt++] = clusterid;
    return 1;
  }
}

int iidcomp(const void *i1, const void *i2)
{
  return *(int *)i1 - *(int *)i2;
}

void CleanSession(SESSION *s)
{
  int i, j;

  if (s->cnt<=1) return;
  qsort(s->ids, s->cnt, sizeof(s->ids[0]), iidcomp);
  for (j=0, i=1; i<s->cnt; i++) {
    if (s->ids[i]>s->ids[j]) {
      if (++j!=i) s->ids[j]=s->ids[i];
    }
  }
  s->cnt=j+1;
}


int groupbycid(const void *v1, const void *v2)
{
  SidCid *t1, *t2;

  t1 = (SidCid *)v1;
  t2 = (SidCid *)v2;
  if (t1->cid > t2->cid)
    return 1;
  else if (t1->cid < t2->cid)
    return -1;
  if (t1->sid > t2->sid)
    return 1;
  else 
    return -1;
}

int orderbysupport(const void *v1, const void *v2)
{
  int *t1, *t2;

  t1 = (int *)v1;
  t2 = (int *)v2;
  if (t1[1] > t2[1])
    return -1;
  else if (t1[1] < t2[1])
    return 1;
  if (t1[0] > t2[0])
    return -1;
  if (t1[0] < t2[0])
    return 1;
  return 0;
}

int orderbyNumSessions(const void *v1, const void *v2)
{
  CF *t1, *t2;

  t1 = (CF *)v1;
  t2 = (CF *)v2;
  if (t1->NumSessions < t2->NumSessions)
    return -1;
  else if (t1->NumSessions > t2->NumSessions)
    return 1;
  return 0;
}

void DumpTopItems()
{
  int i, j, k;
  int *ClusterItems=NULL;
  int item_array_size, item_array_cnt, newsize;
  void *temp;
  char **item_array;
  FILE *itemfile;
  char item_str[2048];

  /* read the item name from itemfile into item_array */
  printf("Reading itemfile ..."); fflush(stdout);
  item_array_size = BLKSIZE;
  item_array_cnt = 0;
  item_array = malloc(sizeof(char *) * item_array_size);
  itemfile = fopen(itemfilename, "r");
  if (!itemfile) error("Error opening item file.");

  while (fscanf(itemfile,"%s\n", item_str) != EOF) {
    if (item_array_cnt == item_array_size) {
      newsize = item_array_size + (item_array_size >> 1);
      assert(newsize > item_array_size);
      temp = realloc(item_array, sizeof(char *)*newsize);
      if (!temp) error("Not enough memory.");
      item_array = (char **)temp;
      item_array_size = newsize;
    }
    item_array[item_array_cnt] = malloc(strlen(item_str) + 1);
    if (!item_array[item_array_cnt]) error("Not enough memory.");
    strcpy(item_array[item_array_cnt],item_str);
    item_array_cnt++;
  }
  fclose(itemfile);
  printf("done.\n");

  qsort(cf, cfcnt, sizeof(CF), orderbyNumSessions);  /* sort cf[] order by NumSessions */

  for (i=0; i<cfcnt; i++) {
    if (cf[i].NumSessions>1) { /* dump detailed item name for non-singleton clusters */
      printf("C%d: N=%d, Large/AllItems=%d/%d\n", 
	     i+1, cf[i].NumSessions, cf[i].NumLarge, cf[i].hash->cnt);
      ClusterItems=calloc(sizeof(int), 2*cf[i].hash->cnt);
      if (!ClusterItems) error("Not enough memory.");
      for (j=0,k=0;j<cf[i].hash->bulksize*BULKS;j++) {
	if (IsPointer(cf[i].hash->bulk[j])) {
	  ClusterItems[k++]=cf[i].hash->bulk[j]->iid;
	  ClusterItems[k++]=cf[i].hash->bulk[j]->support;
	  assert(k<=cf[i].hash->cnt*2);
	}
      }
      assert(k==cf[i].hash->cnt*2);
      qsort(ClusterItems, cf[i].hash->cnt, sizeof(int)*2, orderbysupport);
      for (k=0; k<20&&k<2*cf[i].hash->cnt; k+=2) {  /* modify k<20 to show more items */
	printf("\t%s, support=%d\n", 
	       item_array[ClusterItems[k]], ClusterItems[k+1]);
      }
      free(ClusterItems);
    }
  }
  
  /* free item array */
  for (i = 0; i < item_array_cnt; i++)  free(item_array[i]);
  if (item_array)  free(item_array);
}

void SaveResult()
{
  int  i, j, c, d;
  FILE *resultfile, *sessionfile, *itemfile, *clusterfile, *plotfile=NULL;
  SidCid *sidcid;
  SESSION **session_array, *s;      /* store all sessions length >=3  */
  NODE *node;
  char item_str[2048], temp_str[256]; 
  void *temp;
  int  item_array_size, item_array_cnt;  
  int  newsize;       /* used to change the array size dynamically */
  int  lastsid;
  char **item_array;
  int  minsup;

  if (resultcnt <=0 ) {
    printf("No result written.\n");
    return;
  }

  assert(Result);
  sidcid = (SidCid *)malloc(sizeof(SidCid)*resultcnt);
  if (!sidcid) error("Not enough memory allocating sidcid.");

  clusterfile = fopen(clusterfilename, "w");
  if (!clusterfile) error("Error writing cluster file.");
  printf("write the clustering result to %s ...\n", clusterfilename);
  for (i=0; i<resultcnt; i++) {
    sidcid[i].sid = i;
    sidcid[i].cid = Result[i];
    fprintf(clusterfile, "%d\n", Result[i]);
  }
  fclose(clusterfile);
  if (0 == OutputMode)   return;   /* clusterfile is the only output file */

  qsort(sidcid, resultcnt, sizeof(SidCid), groupbycid);

  resultfile = fopen(resultfilename, "w");
  if (!resultfile) error("Error: writing result file.");
  printf("write the detail clustering to %s ...\n", resultfilename);
  printf("session:%d \t cluster:%d\n", resultcnt, cfcnt);
  /* output the result as C0: s0 s1..... */
  c = -1;   j = 0;
  for (i = 0; i<resultcnt; i++) {
    if (c != sidcid[i].cid) {
      c = sidcid[i].cid;
      if (c != 0) fprintf(resultfile, "-- %d -- \n\n", j);
      j = 0;
      fprintf(resultfile, "C%d: \t", c);
    }
    fprintf(resultfile, "s%d  ", sidcid[i].sid);
    j++;
  }

  fprintf(resultfile, "-- %d --\n\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n", j);

  /* read the item name from itemfile into item_array */
  item_array_size = BLKSIZE;
  item_array_cnt = 0;
  item_array = malloc(sizeof(char *) * item_array_size);
  itemfile = fopen(itemfilename, "r");
  if (!itemfile) error("Error: opening item file.");

  while (fscanf(itemfile,"%s\n", item_str) != EOF) {
    if (item_array_cnt == item_array_size) {
      newsize = item_array_size + (item_array_size >> 1);
      assert(newsize > item_array_size);
      temp = realloc(item_array, sizeof(char *)*newsize);
      if (!temp) error("Not enough memory.");
      item_array = (char **)temp;
      item_array_size = newsize;
    }
    item_array[item_array_cnt] = malloc(strlen(item_str) + 1);
    if (!item_array[item_array_cnt]) error("Not enough memory.");
    strcpy(item_array[item_array_cnt],item_str);
    item_array_cnt++;
  }
  fclose(itemfile);

  /* read the session id from sessionfile into session_array */
  sessionfile = fopen(sessionfilename, "rb");
  if (!sessionfile) error("Can't open session file."); 
  
  fread(&TotalSessions, sizeof(TotalSessions), 1, sessionfile);  
  fread(&TotalItems, sizeof(TotalItems), 1, sessionfile);
  if ((TotalItems<=0) || (TotalSessions<=0)) {
    printf("Empty file: sessions %ld, items %ld\n", TotalSessions, TotalItems);
    error("Aborted!");
  }

  s = malloc(sizeof(SESSION));
  if (!s) error ("Not enough memoty.");
  s->cnt = 0;
  s->size = 1;
  session_array = calloc(TotalSessions, sizeof(SESSION *));
  if (!session_array) error("Not enough memory.");

  lastsid = -1;  i =0;
  d = fread(&request, sizeof(long), 2, sessionfile);
  while ((d == 2)&&!feof(sessionfile)&&i<MaxSessions) {
    if (lastsid!=request.sid) { /* new session */
      if (lastsid != -1) {
	CleanSession(s); /* sort and remove dups */
	if (s->cnt>=1) { /* only cluster large sessions */
	  /* add the session s to session_array */
	  session_array[i] = malloc(sizeof(SESSION) + (s->cnt -1)* sizeof(int));
	  if (!session_array[i])  error("Not enough memory.");
	  session_array[i]->size = s->cnt;
	  session_array[i]->cnt = s->cnt;
	  for (j = 0; j < s->cnt; j++)
	    session_array[i]->ids[j] = s->ids[j];
	  i++;	      	     	   
	}
      }
      lastsid=request.sid;
      s->cnt=0;
    }
    /* add iid to s */
    if (s->cnt >= s->size) { /* allow more memory */
      assert(s->cnt == s->size);
      newsize = s->size + ((s->size > BLKSIZE) ? (s->size >> 1) : BLKSIZE);
      assert(newsize > s->size);
      s = realloc(s, sizeof(SESSION) + sizeof(int) * (newsize-1));
      if (!s) error("Not enough memory.");	
      s->size=newsize; 
    }
    s->ids[s->cnt++]=request.iid;
    d = fread(&request, sizeof(long), 2, sessionfile);
  } /* end of while for read session file */
  if (lastsid != -1) {
    CleanSession(s); /* sort and remove dups */
    if (s->cnt>=1) { /* only cluster large sessions */
      /* add the session s to session_array */
      session_array[i] = malloc(sizeof(SESSION) + (s->cnt -1)* sizeof(int));
      if (!session_array[i])  error("Not enough memory.");
      session_array[i]->size = s->cnt;
      session_array[i]->cnt = s->cnt;
      for (j = 0; j < s->cnt; j++)
	session_array[i]->ids[j] = s->ids[j];
      i++;	      	     	   
    }
  }
  fclose(sessionfile);

  /* print the cluster:items to the result file */
  for (i=0; i<cfcnt; i++) {
    minsup =  GetMinSupp(cf[i].NumSessions);
    if (cf[i].head) {
      fprintf(resultfile,"C%d : min support = %d  Total sessions = %d  Total items = %d----------------\n", i, minsup, cf[i].NumSessions, cf[i].hash->cnt);
      node = cf[i].head;
      while (node->next) node = node->next;
      c = 1;   /* used as the counter in plotfile */
      if (2 == OutputMode) {         /* output the plotfile */
	sprintf(temp_str, "%s-%d", plotfilename, i);
	plotfile = fopen(temp_str, "w");
	if (!plotfile)  error("Not enough memory.");
      }
      while (node != NULL) {
	if (node->iid > item_array_cnt) error("Itemfile doesn't match Sessionfile.");
	fprintf(resultfile,"\t( ");
	j = 0;
	while (sidcid[j].cid != i) j++;
	while (sidcid[j].cid == i) {
	  if (IsInSession(session_array[sidcid[j].sid], node->iid)) 
	    fprintf(resultfile,"S%d ", sidcid[j].sid); 
	  j++;
	}	
	fprintf(resultfile,")\t");
	if (node->support >= minsup) 
	  fprintf(resultfile, "\titem=%d***sup=%d  %s\n", node->iid, node->support,
		  item_array[node->iid]);
	else
	  fprintf(resultfile, "\titem=%d : sup=%d  %s\n", node->iid, node->support,
		  item_array[node->iid]);
	if (2 == OutputMode) {
	  fprintf(plotfile, "%d\t%d\t%d\t%d\t%s\n", i, c, node->support, 
		  node->iid, item_array[node->iid]); /* clusterid, #, sup, itemID, itemname */
	}
	node = node->prev;
	c++;
      }
      if (2 == OutputMode)  fclose(plotfile);
    }
  } /* end of for */
  fclose(resultfile);

  /* free the space */
  for (i = 0; i < item_array_cnt; i++)  free(item_array[i]);
  free(item_array);
  for (i = 0; session_array[i] != NULL && i < TotalSessions; i++)  free(session_array[i]);
  free(session_array);
  free(s);

} /* end of SaveResult */

void PrintPercentageBar(char *c, int percent, int total)
{
  int i;

  assert(total>0);
  for (i=0; i<percent*60/total; i++) printf(c);
}

void DumpClusterResult()
{
  int i, Sum, SumLarge, MaxSessions;

  for (i=0, MaxSessions=Sum=SumLarge=0; i<cfcnt; i++) {
    Sum+=cf[i].NumSessions;
    MaxSessions = MaxSessions>cf[i].NumSessions ?
      MaxSessions : cf[i].NumSessions;
    SumLarge+=cf[i].NumLarge;
  }

  for (i=0; i<cfcnt; i++) {
    if (cf[i].hash) {
      printf("%2d|", i);
      PrintPercentageBar("-",cf[i].NumSessions, MaxSessions);
      printf("(%d/%d %0.2f%%)\n", 
	     cf[i].NumSessions, Sum, cf[i].NumSessions*100/(float)Sum);
      printf("  |");
      PrintPercentageBar(".",cf[i].NumLarge, SumLarge);
      printf("(%d/%d)\n", cf[i].NumLarge, cf[i].hash->cnt);
    }
  }
}

int ProcessOneSession(SESSION *s, int pass, int sid)
{
  int MinDelta, OriginalDelta, d, OldCost;
  int BestClusterID, BestClusterSize, i;

  info(INFO_INFO2, "---- Session %d: ", sid);
  if (VerboseLevel>=INFO_INFO2) DumpSession(s); 

  if (cfcnt == 0 || (cf[cfcnt-1].NumSessions!=0 && cfcnt < MaxClusters)) {
    AddNewCluster(); 
  }  
  
  BestClusterID = -1;
  MinDelta = OriginalDelta = BestClusterSize = 0;
  OldCost=Cost();
  if (pass > 0) { /* not the first pass, first remove session */
    /* have to perform remove, else LargeHash/SmallHash will be incorrect in the following */
    /* ComputeDeltaCost() calls */
    RemoveSession(Result[sid], s); 
  }
  for (i=0; i<cfcnt; i++) {
    info(INFO_DEBUG, "Testing cluster %d ...\n", i);
    d = ComputeDeltaCost(DELTA_ADD, i, s); /* how many will cost change */
    if (pass>0 && (i == Result[sid])) {
      OriginalDelta=d;
      if (VerboseLevel==INFO_WARN) printf("*%d ", d);
    } else if (VerboseLevel==INFO_WARN) printf("%d ", d);
    if (BestClusterID==-1 || MinDelta > d 
	|| (MinDelta==d && BestClusterSize > cf[i].NumSessions)) {
      MinDelta = d;
      BestClusterID = i;
      BestClusterSize = cf[i].NumSessions;
    }
  }
  assert(BestClusterID!=-1);
  if (pass==0) {
    AddSession(BestClusterID, s);
    info(INFO_WARN, "| %d => %d[%d]\n", 
	 sid, BestClusterID, cf[BestClusterID].NumSessions);
  } else {
    if (MinDelta!=OriginalDelta 
	|| (MinDelta==OriginalDelta && BestClusterSize < cf[Result[sid]].NumSessions)) {
      AddSession(BestClusterID, s);
      info(INFO_WARN, "| %d: %d=>%d[%d] (%d->%d,%d,%d)\n", 
	   sid, Result[sid], BestClusterID, cf[BestClusterID].NumSessions, 
	   OldCost, Cost(), MinDelta, OriginalDelta);
    } else {
      BestClusterID = Result[sid];
      AddSession(Result[sid], s);
      info(INFO_WARN, "| %d: %d[%d] unchanged\n", 
	   sid, Result[sid], cf[Result[sid]].NumSessions);
    }
  }
  return(SetResult(sid, BestClusterID));
}

void ProcessSessionFile()
{
  long    requestsread, sessionsread;
  SESSION *s;
  void    *temp; int newsize;
  int     d, i;
  int     lastsid;
  FILE    *sessionfile;
  int     pass, changed;
  unsigned int OldCost=-1;
  NODE    *p;

  sessionfile = fopen(sessionfilename, "rb");
  if (!sessionfile) error("Can't open session file."); 
  
  fread(&TotalSessions, sizeof(TotalSessions), 1, sessionfile);  
  fread(&TotalItems, sizeof(TotalItems), 1, sessionfile);
  if ((TotalItems<=0) || (TotalSessions<=0)) {
    printf("Empty file: sessions %ld, items %ld\n", TotalSessions, TotalItems);
    error("Aborted!");
  }
  
  s = (SESSION *)malloc(sizeof(SESSION));
  if (!s) error("Not enough memory.");
  s->cnt=0; s->size=1;

  pass = 0; 
  while ((pass == 0 || Cost()<OldCost) && (pass < RotateTimes)){
    if (pass>0) OldCost = Cost();

    fseek(sessionfile, sizeof(long)*2, SEEK_SET);
    info(INFO_INFO, "%ld sessions with %ld unique items expected. "
	 "Reading ('.' = 1K sessions) ", 
	 TotalSessions, TotalItems); 
    requestsread=0; sessionsread=0;

    changed = 0;  lastsid = -1;
    d = fread(&request, sizeof(long), 2, sessionfile);
    while ((d == 2)&&!feof(sessionfile)&&sessionsread<MaxSessions) {
      ++requestsread;
      if (lastsid!=request.sid) { /* new session */
	if (lastsid != -1) {
	  CleanSession(s); /* sort and remove dups */
	  if (s->cnt>=1) { /* only cluster large sessions */
	    if (ProcessOneSession(s, pass, sessionsread)) ++changed;
	    if (++sessionsread % (1024) == 0) { 
	      printf("."); 
	      fflush(stdout);
	    }
	  }
	}
	lastsid=request.sid;
	s->cnt=0;
      }
      /* add iid to s */
      if (s->cnt >= s->size) { /* allow more memory */
	assert(s->cnt == s->size);
	newsize = s->size + ((s->size > BLKSIZE) ? (s->size >> 1) : BLKSIZE);
	assert(newsize > s->size);
	temp = realloc(s, sizeof(SESSION) + sizeof(int) * (newsize-1));
	if (!temp) error("Not enough memory.");
	s = temp;
	s->size=newsize; 
      }
      s->ids[s->cnt++]=request.iid;
      d = fread(&request, sizeof(long), 2, sessionfile);
    } /* end of while for read session file */

    if (lastsid != -1) {
      CleanSession(s); /* sort and remove dups */
      if (s->cnt>=1) { /* only cluster large sessions */
	if (ProcessOneSession(s, pass, sessionsread)) ++changed;
	if (++sessionsread % (1024) == 0) { 
	  printf("."); 
	  fflush(stdout);
	}
      }
    }

    if (sessionsread != TotalSessions && sessionsread < MaxSessions) 
      printf("Warning: sessionsread(%ld) != expected sessions(%ld).\n",
	     sessionsread, TotalSessions);
    ++pass;
    info(INFO_INFO, "====== Pass: %d  Changed: %d  Cost: %d ==========\n", 
	 pass, changed, Cost());
  } /* end of while for pass */
  
  fclose(sessionfile);
  free(s);
  DumpClusterResult();
  SaveResult();

  printf("pass=%d\n", pass);

  /* free the allocated space */
  if (Result)  free(Result);
  for(i = 0; i < cfcnt; i++) {
    if (cf[i].hash)  free(cf[i].hash);
    if (cf[i].head) {
      p = cf[i].head;
      while (p->next) {
	p = p->next;
	free(p->prev);
      }
      free(p);
    }
  }
  if (cf)  free(cf);
  if (LargeHash)  free(LargeHash);
  if (SmallHash)  free(SmallHash);

}

void PrintUsage() 
{
  printf("\n");
  printf("Usage: largeitem DataFile SmallWeight OutputMode [ MinSupport [ MaxClusters [RotateTimes [MaxSessions [VerboseLevel]]]]]\n");
  printf("DataFile is the prefix of sessionfile, itemfile and resultfile\n");
  printf("          sessionfile -- DataFile.session\n");
  printf("          itemfile    -- DataFile.item\n");
  printf("          outputfile  -- DataFile.li_cluster...\n");
  printf("                         DataFile.li_result...\n");
  printf("                         DataFile.li_plot...\n");
  printf("SmallWeight is the weight (default: 1) of smallitem (when the LargeWeight is 1)\n");
  printf("OutputMode can be 0, 1, and 2\n");
  printf("          0 -- output cluster file\n");
  printf("          1 -- output cluster file, and result file\n");
  printf("          2 -- output cluster file, result file, and plot file\n");
  printf("MinSupport is the minimum support for large items (default: 0.6)\n");
  printf("MaxClusters is the maximum cluster number (default: 50)\n");
  printf("RotateTimes is the times of fining process (default: 100)\n");
  printf("MaxSessions is the maximal number of sessions to process (default: 1M)\n");
  printf("\n");
}

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
  printf("Large Item Cluster Algorithm.\n%s\n", vcid);
  if ((argc < 3) || (argc > 9)) {                        /* parameter count is not right */
    PrintUsage();
    exit(1);
  }
  if (argc > 8)  VerboseLevel = atoi(argv[8]);
  if (argc > 7)  MaxSessions = atoi(argv[7]);
  if (argc > 6)  RotateTimes = atoi(argv[6]);
  if (argc > 5)  MaxClusters = atoi(argv[5]);
  if (argc > 4)  MinSupport = atof(argv[4]);
  if (MinSupport>0.99999 || MinSupport < 0.000001) {
    printf("Bad MinSupport (0<MinSupport<1): %f\n", MinSupport);
    exit(1);
  }
  OutputMode = atoi(argv[3]);
  if ( (0 != OutputMode) && (1 != OutputMode) && (2 != OutputMode)) {
    printf("OutputMode value %d is not valid.\n", OutputMode);
    PrintUsage();
    exit(1);
  }
  SMALLWEIGHT = atoi(argv[2]);
  sprintf(sessionfilename, "%s.session", argv[1]);
  sprintf(itemfilename, "%s.item", argv[1]);
  sprintf(resultfilename, "%s.li_result.%.3f", argv[1], MinSupport);
  sprintf(clusterfilename, "%s.li_cluster.%.3f", argv[1], MinSupport);
  sprintf(plotfilename, "%s.li_plot.%.3f", argv[1], MinSupport);    

  info(INFO_INFO, "Session file: %s\n", sessionfilename);
  info(INFO_INFO, "Item file:    %s\n", itemfilename);
  info(INFO_INFO, "Cluster file: %s\n", clusterfilename);
  info(INFO_INFO, "Result file:  %s\n", resultfilename);
  info(INFO_INFO, "Parameter:    %.3f\n", MinSupport);
  info(INFO_INFO, "OutputMode:   %d\n", OutputMode);
  info(INFO_INFO, "LargeWeight:  %d\n", LARGEWEIGHT);
  info(INFO_INFO, "SmallWeight:  %d\n", SMALLWEIGHT);
  info(INFO_INFO, "MaxCluster:   %d\n", MaxClusters);
  info(INFO_INFO, "RotateTimes:  %d\n", RotateTimes);
  info(INFO_INFO, "MaxSessions:  %d\n", MaxSessions);
  info(INFO_INFO, "VerboseLevel: %d\n", VerboseLevel);

  LargeHash = NewHash(MINBULKSIZE);
  SmallHash = NewHash(MINBULKSIZE);

  ProcessSessionFile();
  
  return 0;
}
